#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/if_team.h>

struct wrr_priv {
   uint32_t sent_packets;
   int16_t next_port;
   int16_t max_weight;
   int16_t current_weight;
   int16_t gdc_result;
 };
static struct wrr_priv *get_wrr_priv(struct team *team)
 {
    return (struct wrr_priv *) &team->mode_priv;
 }

static int wrr_set_max_weight(struct team *team){
    int port_index;
    int max_weight = 0;
    struct team_port *port;

    for (port_index = 0; port_index < team->en_port_count; port_index++) {
        port = team_get_port_by_index_rcu(team, port_index);
        if (port->priority > max_weight && team_port_txable(port))
             max_weight = port->priority;
    }
    return max_weight;
}

static bool wrr_transmit(struct team *team, struct sk_buff *skb)
{
    struct wrr_priv *wrr_priv = get_wrr_priv(team);
    struct team_port *port;
    while(true){
        wrr_priv->max_weight = wrr_set_max_weight(team);
        if ( wrr_priv->max_weight <= 0 ) {
            goto drop;
        }
        if (wrr_priv->current_weight > wrr_priv->max_weight) {
            wrr_priv->current_weight = wrr_priv->max_weight;
        }
        wrr_priv->next_port = (wrr_priv->next_port +1) % team->en_port_count;


        if ( wrr_priv->next_port == 0){
            wrr_priv->current_weight = wrr_priv->current_weight - wrr_priv->gdc_result;
            if (wrr_priv->current_weight <= 0){
                wrr_priv->current_weight = wrr_priv->max_weight;
            }
            if (wrr_priv->current_weight == 0){
                goto drop;
            }
        }

        port = team_get_port_by_index_rcu(team,wrr_priv->next_port);

        if ( port->priority >= wrr_priv->current_weight){
                if (team_port_txable(port)) {
                    if (team_dev_queue_xmit(team, port, skb))
                        return false;
                    return true;
                }
        }

    }
drop:
    dev_kfree_skb_any(skb);
    printk(KERN_INFO "dropping packets");
    return false;
}

static int wrr_init(struct team *team){

    struct wrr_priv *wrr_priv = get_wrr_priv(team);
    wrr_priv->gdc_result = 1;
    wrr_priv->max_weight = wrr_set_max_weight(team);
    wrr_priv->next_port = -1;
    wrr_priv->current_weight = 0;

    return 0;

}

 static const struct team_mode_ops wrr_mode_ops = {
          .init                   = wrr_init,
          .transmit               = wrr_transmit,
          .port_enter             = team_modeop_port_enter,
          .port_change_dev_addr   = team_modeop_port_change_dev_addr,
  };

  static const struct team_mode wrr_mode = {
          .kind           = "testing",
          .owner          = THIS_MODULE,
          .priv_size      = sizeof(struct wrr_priv),
          .ops            = &wrr_mode_ops
  };

 static int __init wrr_init_module(void)
  {
	int errs = team_mode_register(&wrr_mode);
        printk(KERN_INFO "result is %d", errs);
	return errs;
  }

  static void __exit wrr_cleanup_module(void)
  {
          team_mode_unregister(&wrr_mode);
  }

  module_init(wrr_init_module);
  module_exit(wrr_cleanup_module);

  MODULE_LICENSE("GPL v2");
  MODULE_AUTHOR("John Hodson <john.hodson@evolving.net.uk>");
  MODULE_DESCRIPTION("weighted Round-robin mode for team");
  MODULE_ALIAS("team-mode-testing");
